#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <cuda.h>
#include <gsl/gsl_cdf.h>
#include "Chi8.h"


int main(int argc ,char* argv[]) {
  FILE *fp;
  size_t size;
	  
/* Initialize rows, cols, ncases, ncontrols from the user */
  unsigned int rows=atoi(argv[2]);
  unsigned int cols=atoi(argv[3]);
  int ncases=atoi(argv[4]);
  int ncontrols=atoi(argv[5]);
  float bcutoff = atof(argv[6]);
  int CUDA_DEVICE = atoi(argv[7]);
  printf("Individuals=%d SNPs=%d cases=%d controls=%d Bonferroni-cutoff=%f CUDA_DEVICE=%d\n",rows,cols,ncases,ncontrols,bcutoff,CUDA_DEVICE);

  cudaError err = cudaSetDevice(CUDA_DEVICE);
  if(err != cudaSuccess) { printf("Error setting CUDA DEVICE\n"); exit(EXIT_FAILURE); }

/*Host variable declaration */

   int THREADS = 256;
   int BLOCKS;
//moved to header file
//   int LD_width = 10;
   float host_results[cols];
   int host_colid[cols];
   int jobs, ref_col;
   int transfer = 0;
   double pair_cutoff = 0;
   unsigned long totaljobs;
   unsigned long totaljobs_LD;
   unsigned long ulone = 1;
   unsigned long ultwo = 2; 
   int final_colid;
   float result; int colid;
   double p_temp;

/*Kernel variable declaration */

   unsigned char *dev_dataT;
   float *results;
   int *dev_colid;

/* Variable initialized to perform univariate tests */

float c_zero, c_one, c_two, con_two,con_zero,con_one,c1_expected,c2_expected, c3_expected,c_total;
float con1_expected, con2_expected, con3_expected,con_total ;
float chi,total,t_col1,t_col2,t_col3;
float p[cols];
float skip[cols];
int skippos[cols];
int skipcount = 0;

/* set the cut off */

float cutoff = 0.05/cols;
printf("Univariate cutoff is %e\n",cutoff);

/* Validation to check if the data file is readable */
fp = fopen(argv[1], "r");
if (fp == NULL) {
    	printf("Cannot Open the File");
	return 0;
}

size = (size_t)(rows * cols);
totaljobs = cols * (cols - ulone) / ultwo ;
totaljobs_LD = (cols - LD_width) * (cols - LD_width - ulone) / ultwo ;
//For chi-square 8-df Bonferroni correction
pair_cutoff = bcutoff / totaljobs_LD ;
//For Pearson correlation
//pair_cutoff = bcutoff ;
printf("Totaljobs is %lu\n",totaljobs);
printf("Size of the data = %u\n",size);
printf("Pair_cutoff = %e\n",pair_cutoff);

fflush(stdout);

unsigned char *dataT = (unsigned char*)malloc((size_t)size);
if(dataT != NULL) {


/* Transfer the SNP Data from the file to CPU Memory */
for(unsigned int i=0 ; i < size; i++ ) {
	int tmp;
	   
	fscanf(fp,"%d",&tmp);
	   
	dataT[i] = (char)(((int)'0')+tmp);
	if(i == size - 1){ 
		transfer = 1; 
		printf("SNP Data Transferred to the Memory....\nProcessing\n");
   	}
} 
fclose(fp) ;

/* Univariate tests on all columns */

for (unsigned int m = 0 ; m < cols ; m++ ) {
     unsigned int n = m;
     c_one = 1.0f; c_zero = 1.0f; c_two = 1.0f; con_one = 1.0f ; con_zero = 1.0f;con_two = 1.0f;
      while(n < ncases * cols) {
           if(dataT[n] == '0') { c_zero ++; }
            	else if(dataT[n] == '1') { c_one++; }
                else if(dataT[n] == '2') { c_two++ ; }
                n = n + cols;
           }
      c_total = c_zero + c_one + c_two;
      n = m + ncases * cols;
      while(n < size) {
           if(dataT[n] == '0') { con_zero++; }
               	else if(dataT[n] == '1') { con_one++;  }
                else if(dataT[n] == '2') { con_two++ ; }
                n = n + cols;
            }
      con_total = con_zero + con_one + con_two;
      total = c_total + con_total;
      t_col1 = c_zero + con_zero;
      t_col2 = c_one + con_one;
      t_col3 = c_two + con_two;
      c1_expected = t_col1 * c_total / total;
      c2_expected = t_col2 * c_total / total ;
      c3_expected = t_col3 * c_total / total ;
      con1_expected = t_col1 * con_total / total ;
      con2_expected = t_col2 * con_total / total ;
      con3_expected = t_col3 * con_total / total ;
      chi = (c_zero - c1_expected) * (c_zero - c1_expected) / c1_expected;
      chi += (c_one - c2_expected) * (c_one - c2_expected) / c2_expected;
      chi += (c_two - c3_expected) * (c_two - c3_expected) / c3_expected;
      chi += (con_zero -con1_expected) * (con_zero -con1_expected) / con1_expected;
      chi += (con_one - con2_expected) * (con_one - con2_expected) / con2_expected;
      chi += (con_two - con3_expected) * (con_two - con3_expected) / con3_expected;
       //chisq[m] = chi;
       p[m] = gsl_cdf_chisq_Q(chi,2);
       }
	
if(transfer == 1){

	/* allocate the Memory in the GPU for SNP data */	   
        cudaMalloc((unsigned char**) &dev_dataT, size * sizeof(unsigned char) );
        cudaMalloc((float**) &results, cols * sizeof(float) );
        cudaMalloc((int**) &dev_colid,cols * sizeof(int) ); 

	/*Copy the SNP data to GPU */
        cudaMemcpy(dev_dataT, dataT, size * sizeof(unsigned char), cudaMemcpyHostToDevice);
	cudaMemcpy(dev_colid,host_colid, cols * sizeof(int), cudaMemcpyHostToDevice);

        printf("SNP 1   Uni p-value     SNP 2   Uni p-value     Chi-Square      P-value\n");
        printf("#####   ###########     #####   ###########     ########        ########\n");

        for(int j=0;j < (cols - (2 * LD_width + 1));j++) {
//Uncomment if you want to consider only insignificant pairwise SNPs
//            if( p[j] > cutoff ) { 
//Uncomment if you want to consider all pairwise SNPs
            if( 1 ) { 
		jobs = cols - j;
                ref_col = j;
                BLOCKS = (jobs + THREADS - 1)/THREADS;

	 	/*Calling the kernel function */
		kernel<<<BLOCKS,THREADS>>>(rows,cols,ncases,ncontrols,jobs,ref_col,dev_dataT,results,dev_colid);
		
                /*Copy the results back in host*/
		cudaMemcpy(host_results,results,cols * sizeof(float),cudaMemcpyDeviceToHost);
		cudaMemcpy(host_colid,dev_colid,cols * sizeof(int),cudaMemcpyDeviceToHost);
		
		/* Identify the maximum significant SNP pairs*/
//		for(int k = 11 ; k < jobs; k++) {
		for(int k = LD_width+1 ; k < jobs; k++) {
		        float max = 0.0f ;
//		 	for( int i=11; i < jobs; i++ ){
		 	for( int i=LD_width+1; i < jobs; i++ ){
//Uncomment if you want to consider only insignificant pairwise SNPs
//					if((host_results[i] >= max) && (p[ref_col + host_colid[i]] > cutoff) &&  p[j] > cutoff) {
//Uncomment if you want to consider all pairwise SNPs
					if(host_results[i] >= max) {
						max = host_results[i];
				 		result = host_results[i];
						colid = host_colid[i];
					}	
			}
			host_results[colid] = 0;
//For chi-square 8-df
			p_temp = gsl_cdf_chisq_Q(result,8);
//For Pearson correlation
//			p_temp = result;
			final_colid= ref_col + colid;
//For chi-square 8-df
			if( p_temp < pair_cutoff ) {
//For Pearson correlation
//			if( p_temp > pair_cutoff ) {
				printf("%d	%e	%d	%e	%e	%e\n",ref_col,p[ref_col],final_colid,p[final_colid],result,p_temp);
			}		
		  	else {
				break;
		 	 } 
		}
               }  		
	else {
		skip[skipcount] = p[j];
		skippos[skipcount] = j;
		skipcount++;
		continue;		
	}	
	}

/* free the Memory in the GPU */
//printf("\n###DONE###\n");
//The output here is not correct
//printf("\n**UNIVARIATE SIGNIFICANT SNPS**\n");
//printf("##############################\n");
//printf("SNP POSITION      P-VALUE\n");
//printf("#############     #########\n");
//for(int s=0; s < skipcount; s++){
//	printf("%d        %e\n",skippos[s],skip[s]);
//}
cudaFree( dev_dataT );
cudaFree( results );
cudaFree(dev_colid );
return 0;
}
else {

        printf("ERROR: Complete data is not transferred to the Memory.\n");

}
}
else {
     	
	printf("ERROR: Loading the data to the Memory.\n");

}
}
